#include <stdio.h>
#include <windows.h>
#include <winuser.h>
#include <tcl.h>
#include <tk.h>
#include <tkPlatDecls.h>
#include <tclPlatDecls.h>
#include "tcltools.hxx"
#define NUMBER_OF_OBJECT_VARS 20
static Tcl_Obj *varPtr[NUMBER_OF_OBJECT_VARS];
static int GetVariableIndex(Tcl_Interp *interp, char *string, int *indexPtr);
static void  SetVarToObj _ANSI_ARGS_((int varIndex,Tcl_Obj *objPtr));
static int CheckIfVarUnset(Tcl_Interp *interp, int varIndex);

static int WinMetrics  (ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
static int WinSysInfo  (ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
static int topmostObjCmd  (ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
static int typeObjCmd  (ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
static int TestobjCmd  (ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
static int ntSetPriority  (ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
static int InfoObjTypeCmd  (ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
static int updateEventObjCmd  (ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
static int Tcl_TryObjCmd  (ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
static int ntFreeSpace  (ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
static int dirname  (ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
int Trycatch_Init(Tcl_Interp *interp);
Tcl_Obj *tclGetVarFromObj (Tcl_Interp *interp, Tcl_Obj *obPtr, int createOK);
int createTkCommands(Tcl_Interp *interp)
{
   Tcl_CreateObjCommand(interp, "tkwintools::metrics", WinMetrics,        (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
   Tcl_CreateObjCommand(interp, "tkwintools::sysinfo", WinSysInfo,        (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
   Tcl_CreateObjCommand(interp, "tkwintools::topmost", topmostObjCmd,     (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
   Tcl_CreateObjCommand(interp, "object",              TestobjCmd,        (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
   Tcl_CreateObjCommand(interp, "ntSetPriority",       ntSetPriority,     (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
   Tcl_CreateObjCommand(interp, "updateEvents",        updateEventObjCmd, (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
   Tcl_CreateObjCommand(interp, "ntFreeSpace",         ntFreeSpace,       (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
   Tcl_CreateObjCommand(interp, "infoObj",             InfoObjTypeCmd,    (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
   //Tcl_CreateObjCommand(interp, "try",                 Tcl_TryObjCmd,     (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
   Tcl_CreateObjCommand(interp, "dirname",             dirname,           (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
   Trycatch_Init(interp);
   return 1;
}
static int InfoObjTypeCmd (ClientData dummy,
                           Tcl_Interp *interp,
                           int objc,
                           Tcl_Obj *CONST objv[])
{
   Tcl_Obj *listPtr;
   int offset = -1, i;
   char *opt;
   listPtr = Tcl_NewListObj(0, (Tcl_Obj **) NULL);
   Tcl_IncrRefCount(listPtr);
    /* List all types */
   if (objc == offset+2)
   {
      Tcl_AppendAllObjTypes(interp, listPtr);
    /* Return latest type and if valid bytes */
   }
   else if (objc == offset+3)
   {
      Tcl_ListObjAppendElement(interp, listPtr, Tcl_NewStringObj(
                                                                (objv[offset+2]->typePtr != NULL) ? objv[offset+2]->typePtr->name : "none", -1));
      Tcl_ListObjAppendElement(interp, listPtr,
                               Tcl_NewStringObj((objv[offset+2]->bytes != NULL) ? "yes" : "no", -1));
    /* assume test against remaing args */
   }
   else
   {
      for (i=offset+3; i < objc; i++ )
      {
         opt = Tcl_GetStringFromObj(objv[i],NULL);
         if (objv[offset+2]->typePtr== NULL && strcmp( opt, "none" )==0  ||
             objv[offset+2]->typePtr!= NULL && strcmp( opt, objv[offset+2]->typePtr->name )==0)
         {
            Tcl_SetObjResult(interp, Tcl_NewBooleanObj(1));
            Tcl_DecrRefCount(listPtr);
            return TCL_OK;
         }
      }
      Tcl_SetObjResult(interp, Tcl_NewBooleanObj(0));
      Tcl_DecrRefCount(listPtr);
      return TCL_OK;
   }
   Tcl_SetObjResult(interp, listPtr);
   Tcl_DecrRefCount(listPtr);
   return TCL_OK;
}
/*
 * Change the priority of this and all child processes (windows only)
 */
static int ntSetPriority(ClientData clientData,Tcl_Interp *interp,int objc,Tcl_Obj *CONST objv[])
{
   int inPriority;
   int tPriority;
   DWORD setPriority = NORMAL_PRIORITY_CLASS;
   static const char *stPriorities[] = {"low", "normal", "high", "realtime", NULL};
   enum enPriorities
   {
      LOW,   NORMAL,   HIGH,   REALTIME
   };


/* Check for incorrect number of parameters */
   switch (GetPriorityClass(GetCurrentProcess()))
   {
      case HIGH_PRIORITY_CLASS:
         Tcl_AppendResult(interp,"high",(char *) NULL);
         break;
      case IDLE_PRIORITY_CLASS:
         Tcl_AppendResult(interp,"low",(char *) NULL);
         break;
      case NORMAL_PRIORITY_CLASS:
         Tcl_AppendResult(interp,"normal",(char *) NULL);
         break;
      case REALTIME_PRIORITY_CLASS:
         Tcl_AppendResult(interp,"realtime",(char *) NULL);
         break;
      default:
         Tcl_AppendResult(interp,"unknown",(char *) NULL);
         break;
   }
   if (objc < 2)
   {
      return TCL_OK;
   }                                             /* end if() */
   /**
    * User wants to set the priority
    *
    */
   if ( Tcl_GetIndexFromObj(interp, objv[1], stPriorities, "option",(int)NULL, &inPriority) == TCL_ERROR )
   {
      return(TCL_ERROR);
   }
   switch (inPriority)
   {
      case LOW:
         setPriority = IDLE_PRIORITY_CLASS;
         tPriority = THREAD_PRIORITY_IDLE;
         break;
      case NORMAL:
         setPriority = NORMAL_PRIORITY_CLASS;
         tPriority = THREAD_PRIORITY_NORMAL;
         break;
      case HIGH:
         setPriority = HIGH_PRIORITY_CLASS;
         tPriority = THREAD_PRIORITY_ABOVE_NORMAL;
         break;
      case REALTIME:
         setPriority = REALTIME_PRIORITY_CLASS;
         tPriority = THREAD_PRIORITY_TIME_CRITICAL;
         break;
      default:
         setPriority = NORMAL_PRIORITY_CLASS;
         tPriority = THREAD_PRIORITY_NORMAL;

   }
   /*
    * Set it for this task
    */
   if (SetPriorityClass(GetCurrentProcess(),setPriority) == 0)
   {
      LPVOID lpMsgBuf;
      FormatMessage(
                   FORMAT_MESSAGE_ALLOCATE_BUFFER |
                   FORMAT_MESSAGE_FROM_SYSTEM |
                   FORMAT_MESSAGE_IGNORE_INSERTS,
                   NULL,
                   GetLastError(),
                   MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),   // Default language
                   (LPTSTR) &lpMsgBuf,
                   0,
                   NULL
                   );
      Tcl_ResetResult(interp);
      Tcl_AppendResult(interp,(char *)lpMsgBuf,(char *) NULL);
      LocalFree( lpMsgBuf );
      return TCL_ERROR;
   }
   /*
    * Set it for this thread (may not be needed on single thread)
    */
   else if (SetThreadPriority(GetCurrentThread(),tPriority) == 0)
   {
      LPVOID lpMsgBuf;
      FormatMessage(
                   FORMAT_MESSAGE_ALLOCATE_BUFFER |
                   FORMAT_MESSAGE_FROM_SYSTEM |
                   FORMAT_MESSAGE_IGNORE_INSERTS,
                   NULL,
                   GetLastError(),
                   MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),   // Default language
                   (LPTSTR) &lpMsgBuf,
                   0,
                   NULL
                   );
      Tcl_ResetResult(interp);
      Tcl_AppendResult(interp,(char *)lpMsgBuf,(char *) NULL);
      LocalFree( lpMsgBuf );
      return TCL_ERROR;
   }
   else
   {
      return TCL_OK;
   }
}
/*
 *----------------------------------------------------------------------
 *
 * WinMetrics --
 * Return some windows metrics
 *
 *
 * Results:
 * A standard Tcl result.
 *
 * Side effects:
 * See the user documentation.
 *
 *----------------------------------------------------------------------
 Const SM_CXSCREEN = 0 'X Size of screen
Const SM_CYSCREEN = 1 'Y Size of Screen
Const SM_CXVSCROLL = 2 'X Size of arrow in vertical scroll bar.
Const SM_CYHSCROLL = 3 'Y Size of arrow in horizontal scroll bar
Const SM_CYCAPTION = 4 'Height of windows caption
Const SM_CXBORDER = 5 'Width of no-sizable borders
Const SM_CYBORDER = 6 'Height of non-sizable borders
Const SM_CXDLGFRAME = 7 'Width of dialog box borders
Const SM_CYDLGFRAME = 8 'Height of dialog box borders
Const SM_CYHTHUMB = 9 'Height of scroll box on horizontal scroll bar
Const SM_CXHTHUMB = 10 ' Width of scroll box on horizontal scroll bar
Const SM_CXICON = 11 'Width of standard icon
Const SM_CYICON = 12 'Height of standard icon
Const SM_CXCURSOR = 13 'Width of standard cursor
Const SM_CYCURSOR = 14 'Height of standard cursor
Const SM_CYMENU = 15 'Height of menu
Const SM_CXFULLSCREEN = 16 'Width of client area of maximized window
Const SM_CYFULLSCREEN = 17 'Height of client area of maximized window
Const SM_CYKANJIWINDOW = 18 'Height of Kanji window
Const SM_MOUSEPRESENT = 19 'True is a mouse is present
Const SM_CYVSCROLL = 20 'Height of arrow in vertical scroll bar
Const SM_CXHSCROLL = 21 'Width of arrow in vertical scroll bar
Const SM_DEBUG = 22 'True if deugging version of windows is running
Const SM_SWAPBUTTON = 23 'True if left and right buttons are swapped.
Const SM_CXMIN = 28 'Minimum width of window
Const SM_CYMIN = 29 'Minimum height of window
Const SM_CXSIZE = 30 'Width of title bar bitmaps
Const SM_CYSIZE = 31 'height of title bar bitmaps
Const SM_CXMINTRACK = 34 'Minimum tracking width of window
Const SM_CYMINTRACK = 35 'Minimum tracking height of window
Const SM_CXDOUBLECLK = 36 'double click width
Const SM_CYDOUBLECLK = 37 'double click height
Const SM_CXICONSPACING = 38 'width between desktop icons
Const SM_CYICONSPACING = 39 'height between desktop icons
Const SM_MENUDROPALIGNMENT = 40 'Zero if popup menus are aligned to the left of the memu bar item. True if it is aligned to the right.
Const SM_PENWINDOWS = 41 'The handle of the pen windows DLL if loaded.
Const SM_DBCSENABLED = 42 'True if double byte characteds are enabled
Const SM_CMOUSEBUTTONS = 43 'Number of mouse buttons.
Const SM_CMETRICS = 44 'Number of system metrics
Const SM_CLEANBOOT = 67 'Windows 95 boot mode. 0 = normal, 1 = safe, 2 = safe with network
Const SM_CXMAXIMIZED = 61 'default width of win95 maximised window
Const SM_CXMAXTRACK = 59 'maximum width when resizing win95 windows
Const SM_CXMENUCHECK = 71 'width of menu checkmark bitmap
Const SM_CXMENUSIZE = 54 'width of button on menu bar
Const SM_CXMINIMIZED = 57 'width of rectangle into which minimised windows must fit.
Const SM_CYMAXIMIZED = 62 'default height of win95 maximised window
Const SM_CYMAXTRACK = 60 'maximum width when resizing win95 windows
Const SM_CYMENUCHECK = 72 'height of menu checkmark bitmap
Const SM_CYMENUSIZE = 55 'height of button on menu bar
Const SM_CYMINIMIZED = 58 'height of rectangle into which minimised windows must fit.
Const SM_CYSMCAPTION = 51 'height of windows 95 small caption
Const SM_MIDEASTENABLED = 74 'Hebrw and Arabic enabled for windows 95
Const SM_NETWORK = 63 'bit o is set if a network is present. Const SM_SECURE = 44 'True if security is present on windows 95 system
Const SM_SLOWMACHINE = 73 'true if machine is too slow to run win95.

 */
static int WinMetrics (ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
   int intMetric;
   int intResult;
   Tcl_Obj *resList[2];
   if ( objc != 2 )
   {
      Tcl_WrongNumArgs(interp, 1, objv, "Metric");
      return TCL_ERROR;
   }
   Tcl_GetIntFromObj(interp,objv[1],&intMetric);
   intResult = GetSystemMetrics(intMetric);


   resList[0] = Tcl_NewIntObj(intMetric);
   resList[1] = Tcl_NewIntObj(intResult);

   Tcl_SetObjResult(interp,Tcl_NewListObj (2, resList));
   return TCL_OK;
}

/*
 *
 * Parameter for SystemParametersInfo()


#define SPI_GETBEEP                 1
#define SPI_SETBEEP                 2
#define SPI_GETMOUSE                3
#define SPI_SETMOUSE                4
#define SPI_GETBORDER               5
#define SPI_SETBORDER               6
#define SPI_GETKEYBOARDSPEED       10
#define SPI_SETKEYBOARDSPEED       11
#define SPI_LANGDRIVER             12
#define SPI_ICONHORIZONTALSPACING  13
#define SPI_GETSCREENSAVETIMEOUT   14
#define SPI_SETSCREENSAVETIMEOUT   15
#define SPI_GETSCREENSAVEACTIVE    16
#define SPI_SETSCREENSAVEACTIVE    17
#define SPI_GETGRIDGRANULARITY     18
#define SPI_SETGRIDGRANULARITY     19
#define SPI_SETDESKWALLPAPER       20
#define SPI_SETDESKPATTERN         21
#define SPI_GETKEYBOARDDELAY       22
#define SPI_SETKEYBOARDDELAY       23
#define SPI_ICONVERTICALSPACING    24
#define SPI_GETICONTITLEWRAP       25
#define SPI_SETICONTITLEWRAP       26
#define SPI_GETMENUDROPALIGNMENT   27
#define SPI_SETMENUDROPALIGNMENT   28
#define SPI_SETDOUBLECLKWIDTH      29
#define SPI_SETDOUBLECLKHEIGHT     30
#define SPI_GETICONTITLELOGFONT    31
#define SPI_SETDOUBLECLICKTIME     32
#define SPI_SETMOUSEBUTTONSWAP     33
#define SPI_SETICONTITLELOGFONT    34
#define SPI_GETFASTTASKSWITCH      35
#define SPI_SETFASTTASKSWITCH      36
#if(WINVER >= 0x0400)
#define SPI_SETDRAGFULLWINDOWS     37
#define SPI_GETDRAGFULLWINDOWS     38
#define SPI_GETNONCLIENTMETRICS    41
#define SPI_SETNONCLIENTMETRICS    42
#define SPI_GETMINIMIZEDMETRICS    43
#define SPI_SETMINIMIZEDMETRICS    44
#define SPI_GETICONMETRICS         45
#define SPI_SETICONMETRICS         46
#define SPI_SETWORKAREA            47
#define SPI_GETWORKAREA            48
#define SPI_SETPENWINDOWS          49

#define SPI_GETHIGHCONTRAST        66
#define SPI_SETHIGHCONTRAST        67
#define SPI_GETKEYBOARDPREF        68
#define SPI_SETKEYBOARDPREF        69
#define SPI_GETSCREENREADER        70
#define SPI_SETSCREENREADER        71
#define SPI_GETANIMATION           72
#define SPI_SETANIMATION           73
#define SPI_GETFONTSMOOTHING       74
#define SPI_SETFONTSMOOTHING       75
#define SPI_SETDRAGWIDTH           76
#define SPI_SETDRAGHEIGHT          77
#define SPI_SETHANDHELD            78
#define SPI_GETLOWPOWERTIMEOUT     79
#define SPI_GETPOWEROFFTIMEOUT     80
#define SPI_SETLOWPOWERTIMEOUT     81
#define SPI_SETPOWEROFFTIMEOUT     82
#define SPI_GETLOWPOWERACTIVE      83
#define SPI_GETPOWEROFFACTIVE      84
#define SPI_SETLOWPOWERACTIVE      85
#define SPI_SETPOWEROFFACTIVE      86
#define SPI_SETCURSORS             87
#define SPI_SETICONS               88
#define SPI_GETDEFAULTINPUTLANG    89
#define SPI_SETDEFAULTINPUTLANG    90
#define SPI_SETLANGTOGGLE          91
#define SPI_GETWINDOWSEXTENSION    92
#define SPI_SETMOUSETRAILS         93
#define SPI_GETMOUSETRAILS         94
#define SPI_SETSCREENSAVERRUNNING  97
#define SPI_SCREENSAVERRUNNING     SPI_SETSCREENSAVERRUNNING
#endif  WINVER >= 0x0400
#define SPI_GETFILTERKEYS          50
#define SPI_SETFILTERKEYS          51
#define SPI_GETTOGGLEKEYS          52
#define SPI_SETTOGGLEKEYS          53
#define SPI_GETMOUSEKEYS           54
#define SPI_SETMOUSEKEYS           55
#define SPI_GETSHOWSOUNDS          56
#define SPI_SETSHOWSOUNDS          57
#define SPI_GETSTICKYKEYS          58
#define SPI_SETSTICKYKEYS          59
#define SPI_GETACCESSTIMEOUT       60
#define SPI_SETACCESSTIMEOUT       61
#if(WINVER >= 0x0400)
#define SPI_GETSERIALKEYS          62
#define SPI_SETSERIALKEYS          63
#endif  WINVER >= 0x0400
#define SPI_GETSOUNDSENTRY         64
#define SPI_SETSOUNDSENTRY         65

#if (_WIN32_WINNT >= 0x0400) || (_WIN32_WINDOWS > 0x0400)
#define SPI_GETMOUSEHOVERWIDTH     98
#define SPI_SETMOUSEHOVERWIDTH     99
#define SPI_GETMOUSEHOVERHEIGHT   100
#define SPI_SETMOUSEHOVERHEIGHT   101
#define SPI_GETMOUSEHOVERTIME     102
#define SPI_SETMOUSEHOVERTIME     103
#define SPI_GETWHEELSCROLLLINES   104
#define SPI_SETWHEELSCROLLLINES   105

#define SPI_GETSHOWIMEUI          110
#define SPI_SETSHOWIMEUI          111
#endif


#if(WINVER >= 0x0500)
#define SPI_GETMOUSESPEED         112
#define SPI_SETMOUSESPEED         113
#define SPI_GETSCREENSAVERRUNNING 114
#endif  WINVER >= 0x0500

#if(WINVER >= 0x0500)
#define SPI_GETACTIVEWINDOWTRACKING         0x1000
#define SPI_SETACTIVEWINDOWTRACKING         0x1001
#define SPI_GETMENUANIMATION                0x1002
#define SPI_SETMENUANIMATION                0x1003
#define SPI_GETCOMBOBOXANIMATION            0x1004
#define SPI_SETCOMBOBOXANIMATION            0x1005
#define SPI_GETLISTBOXSMOOTHSCROLLING       0x1006
#define SPI_SETLISTBOXSMOOTHSCROLLING       0x1007
#define SPI_GETGRADIENTCAPTIONS             0x1008
#define SPI_SETGRADIENTCAPTIONS             0x1009
#define SPI_GETMENUUNDERLINES               0x100A
#define SPI_SETMENUUNDERLINES               0x100B
#define SPI_GETACTIVEWNDTRKZORDER           0x100C
#define SPI_SETACTIVEWNDTRKZORDER           0x100D
#define SPI_GETHOTTRACKING                  0x100E
#define SPI_SETHOTTRACKING                  0x100F
#define SPI_GETFOREGROUNDLOCKTIMEOUT        0x2000
#define SPI_SETFOREGROUNDLOCKTIMEOUT        0x2001
#define SPI_GETACTIVEWNDTRKTIMEOUT          0x2002
#define SPI_SETACTIVEWNDTRKTIMEOUT          0x2003
#define SPI_GETFOREGROUNDFLASHCOUNT         0x2004
#define SPI_SETFOREGROUNDFLASHCOUNT         0x2005
#endif  WINVER >= 0x0500

to get monitor number
HMONITOR WINUSERAPI MonitorFromWindow(HWND hwnd,
                                      DWORD dwFlags);
*/
static int WinSysInfo (ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
   int intResult;
   int index;

   //UINT uiAction;                                // system parameter to retrieve or set
   //UINT uiParam;                                 // depends on action to be taken
   //PVOID pvParam;                                // depends on action to be taken
   //UINT fWinIni;                                 // user profile update option
   static const char *optionStrings[] = {
      "-workarea",   "-iconmetrics",  "-monitor",  "-screensaveactivate",  "-deskwallpaper",
      NULL
   };

   enum options
   {
      WORKAREA, EICONMETRICS, MONITORNUM,  SCREENACTIVATE, DESKWALLPAPER
   };

   if (objc < 2)
   {
      Tcl_WrongNumArgs(interp, 1, objv, "Setting ?value?");
      return TCL_ERROR;
   }

   if (Tcl_GetIndexFromObj(interp, objv[1], optionStrings, "options", 0, &index) != TCL_OK)
   {
      return TCL_ERROR;
   }
   switch (index)
   {
      case WORKAREA:
         {
            int i=  0;
            RECT rect;
            Tcl_Obj *resList[4];
            intResult = SystemParametersInfo(SPI_GETWORKAREA,(UINT)NULL,(PVOID)&rect,(UINT)NULL);
            resList[i++] = Tcl_NewIntObj(rect.left);
            resList[i++] = Tcl_NewIntObj(rect.top);
            resList[i++] = Tcl_NewIntObj(rect.right);
            resList[i++] = Tcl_NewIntObj(rect.bottom);
            Tcl_SetObjResult(interp,Tcl_NewListObj (i, resList));
         }
         break;
      case EICONMETRICS:
         {
            int i=  0;
            ICONMETRICS metrics;
            Tcl_Obj *resList[4];
            metrics.cbSize = sizeof(ICONMETRICS);
            intResult = SystemParametersInfo(SPI_GETICONMETRICS,(UINT)NULL,(PVOID)&metrics,(UINT)NULL);
            resList[i++] = Tcl_NewIntObj(metrics.iHorzSpacing);
            resList[i++] = Tcl_NewIntObj(metrics.iVertSpacing);
            resList[i++] = Tcl_NewIntObj(metrics.iTitleWrap);
            Tcl_SetObjResult(interp,Tcl_NewListObj (i, resList));
         }
         break;
      case SCREENACTIVATE:
         {
            BOOL res=FALSE;
            intResult = SystemParametersInfo(SPI_GETLOWPOWERACTIVE,(UINT)NULL,(PVOID)&res,(UINT)NULL);
            if (res)
            {
               Tcl_SetObjResult(interp,Tcl_NewStringObj ("LowPower", -1));
            }
            else
            {
               intResult = SystemParametersInfo(SPI_GETPOWEROFFACTIVE,(UINT)NULL,(PVOID)&res,(UINT)NULL);
               if (res)
               {
                  Tcl_SetObjResult(interp,Tcl_NewStringObj ("PowerOff", -1));
               }
               else
               {
                  Tcl_SetObjResult(interp,Tcl_NewStringObj ("PowerOn", -1));
               }
            }
         }
         break;
   }

   return TCL_OK;
}
/*
 *----------------------------------------------------------------------
 *
 * topmostObjCmd --
 *
 *      This function implements the tk "topmost" command.
 *
 * Results:
 *      A standard Tcl result.
 *
 * Side effects:
 *      Tries to place thew window supplied as an argument at the topmost
 *      level within the application or in the display.
 *
 *----------------------------------------------------------------------
 */

static int
  topmostObjCmd
  (
  ClientData clientData,
  Tcl_Interp *interp,
  int objc,
  Tcl_Obj *CONST objv[]
  )
{
   Tk_Window  tkmain;
   Tk_Window  tkwin;
   char *wpath;
   HWND topw;
   int     ontop    = 1;
   int     was;

   Tcl_ResetResult(interp);
   if (objc < 2 || objc > 3)
   {
      Tcl_WrongNumArgs(interp, 1, objv, "window ?boolean?");
      return TCL_ERROR;
   }
   tkmain = Tk_MainWindow(interp);
   if (tkmain == NULL)
   {
      return TCL_ERROR;
   }
   wpath = Tcl_GetStringFromObj(objv[1], NULL);
   if ((tkwin = Tk_NameToWindow(interp, wpath, tkmain)) == NULL)
   {
      return TCL_ERROR;
   }
   if (!Tk_IsTopLevel(tkwin))
   {
      Tcl_AppendStringsToObj(Tcl_GetObjResult(interp), "window \"", wpath,
                             "\" isn't a top-level window", (char *)NULL);
      return TCL_ERROR;
   }
   if (!Tk_IsMapped(tkwin))
   {
      Tcl_AppendStringsToObj(Tcl_GetObjResult(interp), "window \"", wpath,
                             "\" is not yet mapped", (char *)NULL);
      return TCL_ERROR;
   }
   topw = GetParent(Tk_GetHWND(Tk_WindowId(tkwin)));
   was = (GetWindowLong(topw, GWL_EXSTYLE) & WS_EX_TOPMOST) != 0;
   if (objc == 3)
   {
      if (Tcl_GetBooleanFromObj(interp, objv[2], &ontop) != TCL_OK)
      {
         return TCL_ERROR;
      }
      SetWindowPos(topw, ontop ? HWND_TOPMOST : HWND_NOTOPMOST,
                   0, 0, 0, 0, SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOSIZE);
   }
   Tcl_SetObjResult(interp, Tcl_NewBooleanObj(was));
   return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * updateEventObjCmd --
 *
 *	Granular version of the Update command, can be used
 *	to only allow certain events to occur.
 * e.g. updateEvents timers will allow background timers to fire
 * without allow window updates to occur.
 *      updateEvents windows will only allow other events to occur
 *      if a window event was also processed.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

/* ARGSUSED */
static int
  updateEventObjCmd
  (
  ClientData clientData,
  Tcl_Interp *interp,
  int objc,
  Tcl_Obj *CONST objv[]
  )
{
   int optionIndex;
   int flags = 0;
   static const char *updateOptions[] = {"windows","files","timers","idle","all", (char *) NULL};
   enum updateOptions
   {
      WIN_EVENTS, FILE_EVENTS, TIMER_EVENTS, IDLE_EVENTS,ALL_EVENTS
   };

   if (objc == 1)
   {
      flags = TCL_ALL_EVENTS|TCL_DONT_WAIT;
   }
   else if (objc == 2)
   {
      if (Tcl_GetIndexFromObj(interp, objv[1], updateOptions, "option", 0, &optionIndex) != TCL_OK)
      {
         return TCL_ERROR;
      }
      switch ((enum updateOptions) optionIndex)
      {
         case WIN_EVENTS: {
               flags = TCL_WINDOW_EVENTS|TCL_DONT_WAIT;
               break;
            }
         case FILE_EVENTS: {
               flags = TCL_FILE_EVENTS|TCL_DONT_WAIT;
               break;
            }
         case TIMER_EVENTS: {
               flags = TCL_TIMER_EVENTS|TCL_DONT_WAIT;
               break;
            }
         case IDLE_EVENTS: {
               flags = TCL_IDLE_EVENTS|TCL_DONT_WAIT;
               break;
            }
         case ALL_EVENTS: {
               flags = TCL_ALL_EVENTS|TCL_DONT_WAIT;
               break;
            }
         default: {
               panic("Tcl_UpdateObjCmd: bad option index to updateEvents");
            }
      }
   }
   else
   {
      Tcl_WrongNumArgs(interp, 1, objv, "?windows | files | timers | idle | all?");
      return TCL_ERROR;
   }

   while (Tcl_DoOneEvent(flags) != 0)
   {
      /* Empty loop body */
   }

   /*
    * Must clear the interpreter's result because event handlers could
    * have executed commands.
    */

   Tcl_ResetResult(interp);
   return TCL_OK;
}
/**
 * Determine the freespace available on a drive
 *
 */
static int ntFreeSpace
  (
  ClientData clientData,
  Tcl_Interp *interp,
  int objc,
  Tcl_Obj *CONST objv[]
  )
{
   /* In the Windows API documentation, C string parameters have the type LPCTSTR
      if the string is to be used only as input. (The C in the name indicates
      constant.) Parameters have the type LPTSTR if the string is to be used as
      an output buffer filled by the function. The T in the names indicates that
      the string could be either ANSI (LPCSTR or LPSTR) or Unicode (LPWCSTR or LPWSTR),
      depending on constants passed to the C compiler at compile time.
    */
   int diskbytes;
   int freebytes;
   int tempLen;
   TDOfcl tdo;

   typedef BOOL (WINAPI *P_GDFSE)(LPCTSTR, PULARGE_INTEGER, PULARGE_INTEGER,PULARGE_INTEGER);
   if (objc < 2)
   {
      Tcl_WrongNumArgs(interp, objc, objv,"folder" );
      return TCL_ERROR;
   }                                             /* end if() */
   tdo.Init(interp);

#define UTF
#ifdef UTF
   Tcl_DString ds;
   const char *tempString = Tcl_GetStringFromObj(objv[1],&tempLen);
   const char *folder = Tcl_UtfToExternalDString(NULL,tempString, tempLen, &ds);   // returns char *
#else
   LPCTSTR folder = Tcl_GetStringFromObj(objv[1],&tempLen);
#endif
   P_GDFSE  pGetDiskFreeSpaceEx = (P_GDFSE)GetProcAddress( GetModuleHandle("kernel32.dll"),
                                                           "GetDiskFreeSpaceExA");
   // Note, if folder is unc it should contain a trailing backslash
   if (pGetDiskFreeSpaceEx)
   {
      ULARGE_INTEGER  i64FreeBytesToCaller;      // bytes available to caller
      ULARGE_INTEGER  i64TotalBytes;             // bytes on disk
      ULARGE_INTEGER  i64FreeBytes;              // free bytes on disk
      // LPCTSTR outString = W2CA(folder);
      bool fResult = pGetDiskFreeSpaceEx (folder,
                                          &i64FreeBytesToCaller,
                                          &i64TotalBytes,
                                          &i64FreeBytes);

      // Process GetDiskFreeSpaceEx results.
      if (fResult)
      {
         // Return result in MBytes
         diskbytes = (int)(i64TotalBytes.QuadPart>>20);
         freebytes = (int)(i64FreeBytesToCaller.QuadPart>>20);
         tdo.Add(diskbytes);
         tdo.Add(freebytes);
         tdo.SetResult();
         return TCL_OK;
      }
   }
   else
   {
      DWORD dwSectPerClust;
      DWORD dwBytesPerSect;
      DWORD dwFreeClusters;
      DWORD dwTotalClusters;
      bool fResult = GetDiskFreeSpace (folder,
                                       &dwSectPerClust,
                                       &dwBytesPerSect,
                                       &dwFreeClusters,
                                       &dwTotalClusters);

      if (fResult)
      {
         tdo.Add((int)(dwBytesPerSect*dwSectPerClust*dwTotalClusters>>20));
         tdo.Add((int)(dwBytesPerSect*dwSectPerClust*dwFreeClusters>>20));
         tdo.SetResult();
      }

   }
   tdo.Add("Error Checking");
   tdo.Add((char *)folder);
   tdo.SetResult();
#ifdef UTF
   Tcl_DStringFree(&ds);
#endif
   return TCL_ERROR;
}
/*
 * Return a very fast dirname command
 */
static int dirname
  (
  ClientData clientData,
  Tcl_Interp *interp,
  int objc,
  Tcl_Obj *CONST objv[]
  )
{
   typedef BOOL (WINAPI *P_GDFSE)(LPCTSTR, PULARGE_INTEGER, PULARGE_INTEGER,PULARGE_INTEGER);
   if (objc < 2)
   {
      Tcl_WrongNumArgs(interp, objc, objv,"folder" );
      return TCL_ERROR;
   }                                             /* end if() */

   Tcl_DString ds;
   int tempLen;
   const char *tempString = Tcl_GetStringFromObj(objv[1],&tempLen);
   char *folder = Tcl_UtfToExternalDString(NULL,tempString, tempLen, &ds);   // returns char *

   for (int i=strlen(folder)-1;i>0;i--)
   {
      if (folder[i]=='/')
      {
         folder[i]=NULL;
      }
   }
   Tcl_ResetResult(interp);
   Tcl_AppendToObj(Tcl_GetObjResult(interp), folder, -1);
   Tcl_DStringFree(&ds);
   return TCL_OK;
}


/*
                           F R U  F R U
*/

/*
 *----------------------------------------------------------------------
 *
 * TestobjCmd --
 *
 *      This procedure implements the "testobj" command. It is used to test
 *      the type-independent portions of the Tcl object type implementation.
 *
 * Results:
 *      A standard Tcl object result.
 *
 * Side effects:
 *      Creates and frees objects.
 *
 *----------------------------------------------------------------------
 */

static int
  TestobjCmd
  (
  ClientData clientData,
  Tcl_Interp *interp,
  int objc,
  Tcl_Obj *CONST objv[]
  )
{
   int varIndex, destIndex;
   char *index, *subCmd, *string;
   Tcl_ObjType *targetType;

   if (objc < 2)
   {
      wrongNumArgs:
      Tcl_WrongNumArgs(interp, 1, objv, "option arg ?arg ...?");
      return TCL_ERROR;
   }

   subCmd = Tcl_GetString(objv[1]);
   if (strcmp(subCmd, "assign") == 0)
   {
      if (objc != 4)
      {
         goto wrongNumArgs;
      }
      index = Tcl_GetString(objv[2]);
      if (GetVariableIndex(interp, index, &varIndex) != TCL_OK)
      {
         return TCL_ERROR;
      }
      if (CheckIfVarUnset(interp, varIndex))
      {
         return TCL_ERROR;
      }
      string = Tcl_GetString(objv[3]);
      if (GetVariableIndex(interp, string, &destIndex) != TCL_OK)
      {
         return TCL_ERROR;
      }
      SetVarToObj(destIndex, varPtr[varIndex]);
      Tcl_SetObjResult(interp, varPtr[destIndex]);
   }
   else if (strcmp(subCmd, "convert") == 0)
   {
      char *typeName;
      if (objc != 4)
      {
         goto wrongNumArgs;
      }
      index = Tcl_GetString(objv[2]);
      if (GetVariableIndex(interp, index, &varIndex) != TCL_OK)
      {
         return TCL_ERROR;
      }
      if (CheckIfVarUnset(interp, varIndex))
      {
         return TCL_ERROR;
      }
      typeName = Tcl_GetString(objv[3]);
      if ((targetType = Tcl_GetObjType(typeName)) == NULL)
      {
         Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
                                "no type ", typeName, " found", (char *) NULL);
         return TCL_ERROR;
      }
      if (Tcl_ConvertToType(interp, varPtr[varIndex], targetType)
          != TCL_OK)
      {
         return TCL_ERROR;
      }
      Tcl_SetObjResult(interp, varPtr[varIndex]);
   }
   else if (strcmp(subCmd, "refcount") == 0)
   {
      if (objc != 3)
      {
         goto wrongNumArgs;
      }
      index = Tcl_GetString(objv[2]);
      if (GetVariableIndex(interp, index, &varIndex) != TCL_OK)
      {
         return TCL_ERROR;
      }
      if (CheckIfVarUnset(interp, varIndex))
      {
         return TCL_ERROR;
      }
      Tcl_SetObjResult(interp, Tcl_NewIntObj(varPtr[varIndex]->refCount));
   }
   else if (strcmp(subCmd, "objtype") == 0)
   {
      char *typeName;

   /*
    * return an object containing the name of the argument's type
    * of internal rep.  If none exists, return "none".
    */

      if (objc != 3)
      {
         goto wrongNumArgs;
      }
      if (objv[2]->typePtr == NULL)
      {
         Tcl_SetObjResult(interp, Tcl_NewStringObj("none", -1));
      }
      else
      {
         typeName = objv[2]->typePtr->name;
         Tcl_SetObjResult(interp, Tcl_NewStringObj(typeName, -1));
      }
   }
   else if (strcmp(subCmd, "type") == 0)
   {
      if (objc != 3)
      {
         goto wrongNumArgs;
      }
      index = Tcl_GetString(objv[2]);
      if (GetVariableIndex(interp, index, &varIndex) != TCL_OK)
      {
         return TCL_ERROR;
      }
      if (CheckIfVarUnset(interp, varIndex))
      {
         return TCL_ERROR;
      }
      if (varPtr[varIndex]->typePtr == NULL)
      {                                          /* a string! */
         Tcl_AppendToObj(Tcl_GetObjResult(interp), "string", -1);
      }
      else
      {
         Tcl_AppendToObj(Tcl_GetObjResult(interp),
                         varPtr[varIndex]->typePtr->name, -1);
      }
   }
   else if (strcmp(subCmd, "types") == 0)
   {
      if (objc != 2)
      {
         goto wrongNumArgs;
      }
      if (Tcl_AppendAllObjTypes(interp,
                                Tcl_GetObjResult(interp)) != TCL_OK)
      {
         return TCL_ERROR;
      }
   }
   else
   {
      Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
                             "bad option \"",
                             Tcl_GetString(objv[1]),
                             "\": must be assign, convert, duplicate, freeallvars, ",
                             "newobj, objcount, objtype, refcount, type, or types",
                             (char *) NULL);
      return TCL_ERROR;
   }
   return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * GetVariableIndex --
 *
 *      Utility routine to get a test variable index from the command line.
 *
 * Results:
 *      A standard Tcl object result.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

static int
  GetVariableIndex(Tcl_Interp *interp, char *string, int *indexPtr)
{
   int index;

   if (Tcl_GetInt(interp, string, &index) != TCL_OK)
   {
      return TCL_ERROR;
   }
   if (index < 0 || index >= NUMBER_OF_OBJECT_VARS)
   {
      Tcl_ResetResult(interp);
      Tcl_AppendToObj(Tcl_GetObjResult(interp), "bad variable index", -1);
      return TCL_ERROR;
   }

   *indexPtr = index;
   return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CheckIfVarUnset --
 *
 *      Utility procedure that checks whether a test variable is readable:
 *      i.e., that varPtr[varIndex] is non-NULL.
 *
 * Results:
 *      1 if the test variable is unset (NULL); 0 otherwise.
 *
 * Side effects:
 *      Sets the interpreter result to an error message if the variable is
 *      unset (NULL).
 *
 *----------------------------------------------------------------------
 */

static int
  CheckIfVarUnset(Tcl_Interp *interp, int varIndex)
{
   if (varPtr[varIndex] == NULL)
   {
      char buf[32 + TCL_INTEGER_SPACE];

      sprintf(buf, "variable %d is unset (NULL)", varIndex);
      Tcl_ResetResult(interp);
      Tcl_AppendToObj(Tcl_GetObjResult(interp), buf, -1);
      return 1;
   }
   return 0;
}
/*
 *----------------------------------------------------------------------
 *
 * SetVarToObj --
 *
 *      Utility routine to assign a Tcl_Obj* to a test variable. The
 *      Tcl_Obj* can be NULL.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      This routine handles ref counting details for assignment:
 *      i.e. the old value's ref count must be decremented (if not NULL) and
 *      the new one incremented (also if not NULL).
 *
 *----------------------------------------------------------------------
 */

static void
  SetVarToObj(int varIndex, Tcl_Obj *objPtr)
{
   if (varPtr[varIndex] != NULL)
   {
      Tcl_DecrRefCount(varPtr[varIndex]);
   }
   varPtr[varIndex] = objPtr;
   if (objPtr != NULL)
   {
      Tcl_IncrRefCount(objPtr);
   }
}

/*
  * Implementation of try/catch and throw commands according to TIP 89
  */

#include <tcl.h>

 /*
  * We keep a stack of contexts; whenever we have to handle an error,
  * i.e. are executing a catch {} clause, we store the current error
  * (errorCode, errorInfo and message), so that a throw with no arguments
  * can re-throw it.
  *
  * This is interpreter-specific data. Each element is a list, with the
  * last element being the most current one.
  */

typedef struct
{
   Tcl_Obj * errorCodeStack;
   Tcl_Obj * errorInfoStack;
   Tcl_Obj * errorMsgStack;
   Tcl_Obj * errorCodeName;
   Tcl_Obj * errorInfoName;
} TryCatchTsd;

 /*
  * Throw an Exception
  *
  * throw ?<type> ?<message>? ?<info>??
  *
  * Throws an exception with the errorCode <type>, the message <message>
  * and the errorInfo <info>.
  *
  * An instance of throw with no arguments can be used within a catch or
  * finally block to immediately re-throw the current exception that is
  * being handled by the catch block.
  */

static int
  Tcl_ThrowObjCmd (ClientData clientData, Tcl_Interp *interp,
                   int objc, Tcl_Obj *CONST objv[])
{
   TryCatchTsd * myTsd = (TryCatchTsd *) clientData;

   if (objc < 1 || objc > 4)
   {
      Tcl_AppendResult (interp, "wrong # args: should be \"",
                        Tcl_GetStringFromObj (objv[0], NULL),
                        " ?<type> ?<message>? ?<info>??\"", NULL);
      return TCL_ERROR;
   }

   /*
    * Re-throw an error
    */

   if (objc < 2)
   {
      Tcl_Obj *errorCode, *errorInfo, *errorMsg;
      int lastelement;

      Tcl_ListObjLength (interp, myTsd->errorMsgStack, &lastelement);

      if (lastelement < 1)
      {
         Tcl_AppendResult (interp, "error: throw with no parameters ",
                           "outside of a catch",
                           NULL);
         return TCL_ERROR;
      }

      lastelement--;
      Tcl_ListObjIndex (interp, myTsd->errorMsgStack,
                        lastelement, &errorMsg);
      Tcl_ListObjIndex (interp, myTsd->errorCodeStack,
                        lastelement, &errorCode);
      Tcl_ListObjIndex (interp, myTsd->errorInfoStack,
                        lastelement, &errorInfo);

      Tcl_ResetResult (interp);
      Tcl_SetObjResult (interp, errorMsg);
      Tcl_SetObjErrorCode (interp, errorCode);

#ifdef _TCLINT
      Tcl_ObjSetVar2 (interp, myTsd->errorInfoName, NULL, errorInfo,
                      TCL_GLOBAL_ONLY);
      interp->flags = ERR_IN_PROGRESS;
#else
      Tcl_AddErrorInfo (interp, Tcl_GetStringFromObj (errorInfo, NULL));
#endif
      return TCL_ERROR;
   }

   /*
    * throw with parameters
    */

   Tcl_ResetResult (interp);

   if (objc >= 3)
   {
      Tcl_SetObjResult (interp, objv[2]);
   }
   else
   {
     /*
      * fabricate some error message for human consumption
      */

      Tcl_AppendResult (interp, "error: ",
                        Tcl_GetStringFromObj (objv[1], NULL),
                        NULL);
   }

   Tcl_SetObjErrorCode (interp, objv[1]);

   if (objc >= 4)
   {
#ifdef _TCLINT
      Tcl_ObjSetVar2 (interp, myTsd->errorInfoName, NULL, objv[3],
                      TCL_GLOBAL_ONLY);
      interp->flags = ERR_IN_PROGRESS;
#else
      Tcl_AddErrorInfo (interp, Tcl_GetStringFromObj (objv[3], NULL));
#endif
   }

   /*
    * throw error
    */

   return TCL_ERROR;
}

 /*
  * exception handling
  *
  * try body ?catch {type-list ?ecvar? ?msgvar? ?infovar?} body ...?
  *          ?finally body?
  */

static int
  Tcl_TryObjCmd (ClientData clientData, Tcl_Interp *interp,
                 int objc, Tcl_Obj *CONST objv[])
{
   TryCatchTsd * myTsd = (TryCatchTsd *) clientData;
   int currentIndex, finallyIndex, catchInfoLength, hasCatch;
   char * blockType;
   int res;

   /*
    * first check for syntactic correctness before doing anything
    */

   if (objc < 2)
   {
      Tcl_AppendResult (interp, "wrong # args: should be \"",
                        Tcl_GetStringFromObj (objv[0], NULL),
                        " body ",
                        "?catch {type-list ?ecvar? ?msgvar? ?infovar?} ",
                        "body ...? ",
                        "?finally body?\"", NULL);
      return TCL_ERROR;
   }

   currentIndex = 2;
   finallyIndex = -1;
   hasCatch = 0;

   while (currentIndex < objc)
   {
      blockType = Tcl_GetStringFromObj (objv[currentIndex], NULL);

      if (strcmp (blockType, "catch") == 0)
      {
         Tcl_Obj * typeList;
         int typeListLength;

         if (currentIndex+2 >= objc ||
             Tcl_ListObjLength (interp, objv[currentIndex+1],
                                &catchInfoLength) != TCL_OK ||
             (catchInfoLength < 1 && catchInfoLength > 4) ||
             Tcl_ListObjIndex (interp, objv[currentIndex+1],
                               0, &typeList) != TCL_OK ||
             Tcl_ListObjLength (interp, typeList,
                                &typeListLength) != TCL_OK)
         {
            Tcl_AppendResult (interp, "invalid syntax in catch clause: ",
                              "should be \"",
                              "catch {type-list ?ecvar? ?msgvar? ?infovar?} ",
                              "body\"", NULL);
            return TCL_ERROR;
         }
         hasCatch = 1;
         currentIndex += 3;
      }
      else if (strcmp (blockType, "finally") == 0)
      {
         if (currentIndex+2 != objc)
         {
            Tcl_AppendResult (interp, "trailing args after finally clause",
                              NULL);
            return TCL_ERROR;
         }
         finallyIndex = currentIndex;
         currentIndex += 2;
      }
      else
      {
         Tcl_AppendResult (interp, "invalid syntax: should be \"",
                           Tcl_GetStringFromObj (objv[0], NULL),
                           " body ",
                           "?catch {type-list ?ecvar? ?msgvar? ?infovar?} ",
                           "body ...? ",
                           "?finally body?\"", NULL);
         return TCL_ERROR;
      }
   }

   /*
    * Eval main body
    */

   res = Tcl_EvalObjEx (interp, objv[1], 0);

   /*
    * In case of error, check the catch clauses
    */

   if (res == TCL_ERROR)
   {
      Tcl_Obj *errorCode, *errorInfo, *errorMsg;
      int errorCodeLength, stackLength;

      errorMsg = Tcl_GetObjResult (interp);
      errorCode = Tcl_ObjGetVar2 (interp, myTsd->errorCodeName, NULL,
                                  TCL_GLOBAL_ONLY);
      errorInfo = Tcl_ObjGetVar2 (interp, myTsd->errorInfoName, NULL,
                                  TCL_GLOBAL_ONLY);

     /*
      * After an error has happened, errorCode and errorInfo should
      * exist.
      */

      if (errorCode == NULL || errorInfo == NULL)
      {
         Tcl_AppendResult (interp, "assertion error in try: ",
                           "no errorCode or no errorInfo",
                           NULL);
         return TCL_ERROR;
      }

      if (Tcl_ListObjLength (interp, errorCode, &errorCodeLength) != TCL_OK)
      {
         Tcl_AppendResult (interp, "assertion error in try: "
                           "errorCode is not a list",
                           NULL);
         return TCL_ERROR;
      }

     /*
      * push error data on stack, so that throw can rethrow the error
      */

      Tcl_ListObjAppendElement (interp, myTsd->errorMsgStack, errorMsg);
      Tcl_ListObjAppendElement (interp, myTsd->errorCodeStack, errorCode);
      Tcl_ListObjAppendElement (interp, myTsd->errorInfoStack, errorInfo);

     /*
      * Look for a matching clause
      */

      currentIndex = 2;

      while (currentIndex < objc)
      {
         blockType = Tcl_GetStringFromObj (objv[currentIndex], NULL);

         if (strcmp (blockType, "catch") == 0)
         {
            int typeListLength, matchIndex;
            Tcl_Obj *typeList;

            Tcl_ListObjIndex  (interp, objv[currentIndex+1], 0, &typeList);
            Tcl_ListObjLength (interp, typeList, &typeListLength);

            if (typeListLength > errorCodeLength)
            {
               currentIndex += 3;
               continue;
            }

            for (matchIndex=0; matchIndex<typeListLength; matchIndex++)
            {
               Tcl_Obj *errorCodeItem, *typeListItem;
               const char *errorCodeItemStr, *typeListItemStr;

               Tcl_ListObjIndex (interp, errorCode, matchIndex, &errorCodeItem);
               Tcl_ListObjIndex (interp, typeList, matchIndex, &typeListItem);

               errorCodeItemStr = Tcl_GetStringFromObj (errorCodeItem, NULL);
               typeListItemStr = Tcl_GetStringFromObj (typeListItem, NULL);

               if (!Tcl_StringMatch (errorCodeItemStr, typeListItemStr))
               {
                  break;
               }
            }

            if (matchIndex >= typeListLength)
            {
           /* matching catch clause found */
               break;
            }

         /* continue looking */
            currentIndex += 3;
         }
         else
         {
         /* not a catch clause - there are no matching catch clauses */
            currentIndex = objc;
            break;
         }
      }

     /*
      * Did we find a matching catch clause?
      */

      if (currentIndex < objc)
      {
         Tcl_Obj *ecvar, *msgvar, *infovar;

         Tcl_ListObjLength (interp, objv[currentIndex+1], &catchInfoLength);

       /*
        * set variables with error data
        */

         if (catchInfoLength >= 2)
         {
            Tcl_ListObjIndex (interp, objv[currentIndex+1], 1, &ecvar);
            Tcl_ObjSetVar2 (interp, ecvar, NULL, errorCode, 0);
         }

         if (catchInfoLength >= 3)
         {
            Tcl_ListObjIndex (interp, objv[currentIndex+1], 2, &msgvar);
            Tcl_ObjSetVar2 (interp, msgvar, NULL, errorMsg, 0);
         }

         if (catchInfoLength >= 4)
         {
            Tcl_ListObjIndex (interp, objv[currentIndex+1], 3, &infovar);
            Tcl_ObjSetVar2 (interp, infovar, NULL, errorInfo, 0);
         }

       /*
        * call body; the error code of this body takes precedence
        */

         res = Tcl_EvalObjEx (interp, objv[currentIndex+2], 0);
      }

     /*
      * pop error data from stack
      */

      Tcl_ListObjLength (interp, myTsd->errorMsgStack, &stackLength);
      stackLength--;
      Tcl_ListObjReplace (interp, myTsd->errorMsgStack,
                          stackLength, 1, 0, NULL);
      Tcl_ListObjReplace (interp, myTsd->errorCodeStack,
                          stackLength, 1, 0, NULL);
      Tcl_ListObjReplace (interp, myTsd->errorInfoStack,
                          stackLength, 1, 0, NULL);
   }

   /*
    * Execute finally body. Preserve errorCode and friends; they might
    * be corrupted by the code in the body - e.g. by a try in the code,
    * or in a proc called by the code.
    */

   if (finallyIndex != -1)
   {
      Tcl_Obj *errorCode, *errorInfo, *errorMsg;
      int finallyres, origres=res;

      errorMsg = Tcl_GetObjResult (interp);
      Tcl_IncrRefCount (errorMsg);

      if (origres == TCL_ERROR)
      {
         errorCode = Tcl_ObjGetVar2 (interp, myTsd->errorCodeName, NULL,
                                     TCL_GLOBAL_ONLY);
         errorInfo = Tcl_ObjGetVar2 (interp, myTsd->errorInfoName, NULL,
                                     TCL_GLOBAL_ONLY);
         Tcl_IncrRefCount (errorCode);
         Tcl_IncrRefCount (errorInfo);
      }

      finallyres = Tcl_EvalObjEx (interp, objv[finallyIndex+1], 0);

     /*
      * An Error in the finally clause takes precedence, else restore
      * previous error data
      */

      if (finallyres != TCL_OK)
      {
         res = finallyres;
      }
      else
      {
         Tcl_SetObjResult (interp, errorMsg);

         if (origres == TCL_ERROR)
         {
            Tcl_SetObjErrorCode (interp, errorCode);
#ifdef _TCLINT
            Tcl_ObjSetVar2 (interp, myTsd->errorInfoName, NULL, errorInfo,
                            TCL_GLOBAL_ONLY);
            interp->flags = ERR_IN_PROGRESS;
#else
            Tcl_AddErrorInfo (interp, Tcl_GetStringFromObj (errorInfo, NULL));
#endif
         }
      }

      Tcl_DecrRefCount (errorMsg);

      if (origres == TCL_ERROR)
      { /* bug: these won't be in scope here */
         Tcl_DecrRefCount (errorCode);
         Tcl_DecrRefCount (errorInfo);
      }
   }

   /*
    * Pass along return code
    */

   return res;
}

 /*
  * ----------------------------------------------------------------------
  *
  * "Main" function, install our commands in the Tcl interpreter
  *
  * ----------------------------------------------------------------------
  */

#undef TCL_STORAGE_CLASS
#define TCL_STORAGE_CLASS DLLEXPORT
int
  Trycatch_Init (Tcl_Interp *interp)
{
   TryCatchTsd * myTsd;

#ifdef USE_TCL_STUBS
   if (Tcl_InitStubs (interp, TCL_VERSION, 0) == NULL)
   {
      return TCL_ERROR;
   }
#else
   if (Tcl_PkgRequire (interp, "Tcl", TCL_VERSION, 1) == NULL)
   {
      return TCL_ERROR;
   }
#endif

   /*
    * Allocate Tsd
    */

   myTsd = (TryCatchTsd *) Tcl_Alloc (sizeof (TryCatchTsd));
   myTsd->errorCodeStack = Tcl_NewObj ();
   myTsd->errorInfoStack = Tcl_NewObj ();
   myTsd->errorMsgStack  = Tcl_NewObj ();
   myTsd->errorCodeName  = Tcl_NewStringObj ("errorCode", -1);
   myTsd->errorInfoName  = Tcl_NewStringObj ("errorInfo", -1);

   /*
    * add commands
    */

   Tcl_CreateObjCommand (interp, "throw", Tcl_ThrowObjCmd,
                         (ClientData) myTsd, NULL);
   Tcl_CreateObjCommand (interp, "try", Tcl_TryObjCmd,
                         (ClientData) myTsd, NULL);

   /*
    * Ready
    */

   Tcl_PkgProvide (interp, "trycatch", "0.1");
   return TCL_OK;
}


